package ga;

import ga.core.*;
import ga.listeners.GAListener;

import java.util.Vector;
import java.util.Collections;
import java.util.Random;

/**
 * Runs the GA Algorithm
 *
 * @author Olly Oechsle, University of Essex, Date: 18-May-2007
 * @version 1.0
 */
public class Evolve extends Thread {

    public static long seed = 2357;
    protected static Random r = null;

    public static double getRandomNumber() {

        if (r == null) {
            initialiseRandomNumberGenerator();
        }

        return r.nextDouble();

    }

    public static void initialiseRandomNumberGenerator() {
        if (seed == -1) {
            r = new Random();
        } else {
            r = new Random(seed);
        }
    }

    public void setSeed(long seed) {
        r = null;
        Evolve.seed = seed;
    }

    protected GAProblem problem;
    protected GAParams params;
    protected GAListener listener;

    private GAIndividual bestIndividual;

    public Evolve(GAProblem p, GAListener listener) {
        this.problem = p;
        this.params = p.getParams();
        this.listener = listener;
    }

    public void run() {

        initialiseRandomNumberGenerator();

        listener.initialise(problem);

        listener.onEvolutionStart();

        // Initialise the population
        Vector<GAIndividual> population = new Vector<GAIndividual>(params.getPopulationSize());

        for (int i = 0; i < params.getPopulationSize(); i++) {
            population.add(problem.createIndividual());
        }

        int generation;

        for (generation = 0; generation < params.getGenerations(); generation++) {

            listener.onGenerationStart(generation);

            // Evaluate each individual
            for (int j = 0; j < population.size(); j++) {
                GAIndividual ind = population.elementAt(j);
                if (!ind.evaluated) {
                    problem.evaluate(ind);
                    ind.evaluated = true;
                }
            }

            // Now sort individuals by fitness
            Collections.sort(population);

            // Get the best individual
            bestIndividual = population.elementAt(0);

            // finish early if found ideal individual
            if (bestIndividual.getKozaFitness() == 0) break;

            // Create a new population
            Vector<GAIndividual> offspring = new Vector<GAIndividual>(params.getPopulationSize());

            // Add elites to the new population
            for (int i = 0; i < params.getEliteCount(); i++) {
                offspring.add(population.elementAt(i));
            }

            // Use tournament selection to choose the best individuals
            GATournamentSelection selector = new GATournamentSelection(params, population);

            while (offspring.size() < params.getPopulationSize()) {

                // choose which operator to use for reproduction
                int operator = params.getOperator();

                switch (operator) {
                    case GAParams.CROSSOVER:
                        // get two parents
                        GAIndividual parent1 = selector.select();
                        GAIndividual parent2 = selector.select();
                        GAIndividual[] children = GABreeder.onePointCrossover(problem, parent1, parent2);
                        offspring.add(children[0]);
                        offspring.add(children[1]);
                        break;
                    case GAParams.MUTATION:
                        GAIndividual parent3 = selector.select();
                        GAIndividual child1 = GABreeder.mutation(problem, parent3);
                        offspring.add(child1);
                        break;
                    case GAParams.REPRODUCTION:
                        GAIndividual parent4 = selector.select();
                        GAIndividual child2 = GABreeder.reproduce(problem, parent4);
                        offspring.add(child2);
                        break;
                }

                //GAIndividual parent1 = selector.select();
                //GAIndividual parent2 = selector.select();
                //offspring.add(GABreeder.produceOffspring(problem, parent1, parent2, params));

            }

            if (params.getGenerationGapMethod() == GAParams.GENERATION_GAP_OVERLAP_OFF) {
                // (N,N) evolution
                // children always replace their parents, better or worse
                population = offspring;
                
            } else {

                // (N+N) evolution
                // Children compete with parents to stay in the population

                // Have to evaluate each offspring individual first
                for (int j = 0; j < offspring.size(); j++) {
                    GAIndividual ind = offspring.elementAt(j);
                    if (!ind.evaluated) {
                        problem.evaluate(ind);
                        ind.evaluated = true;
                    }
                }

                Vector<GAIndividual> intermediatePopulation = new Vector<GAIndividual>();
                intermediatePopulation.addAll(population);
                intermediatePopulation.addAll(offspring);

                Collections.sort(intermediatePopulation);

                // only add the best individuals from the combined population
                population.clear();
                for (int i = 0; i < params.getPopulationSize(); i++) {
                    population.add(intermediatePopulation.elementAt(i));
                }

            }

            listener.onGenerationEnd(generation, bestIndividual);

        }

        problem.onFinish(bestIndividual);
        listener.onEvolutionEnd(generation, bestIndividual);

    }

    public GAIndividual getBestIndividual() {
        return bestIndividual;
    }

    public GAParams getParams() {
        return params;
    }
    
}
