package problems.face;

import ga.core.GAIndividual;
import ga.core.GAProblemBoostable;
import ga.classification.BinaryClassifier;

import java.io.File;
import java.util.Vector;

import ac.essex.ooechs.imaging.commons.util.ImageFilenameFilter;
import ac.essex.ooechs.imaging.commons.HaarRegions;
import ac.essex.ooechs.imaging.commons.PixelLoader;
import ac.essex.ooechs.imaging.commons.StatisticsSolver;
import ac.essex.ooechs.adaboost.AdaBoostSample;
import problems.face.util.FaceSample;
import problems.boosting.BoostingProblem;

/**
 * Uses Genetic Algorithms to evolve a simple hypothesis.
 *
 * @author Olly Oechsle, University of Essex, Date: 21-May-2007
 * @version 1.0
 */
public class FaceDetectorProblem1 extends GAProblemBoostable {

    // window has these dimensions
    public static final int WINDOW_WIDTH = 19;
    public static final int WINDOW_HEIGHT = 19;

    // window is divided into blocks
    public static final int WINDOWBLOCKSX = 19;
    public static final int WINDOWBLOCKSY = 19;

    public static final int FACE = BinaryClassifier.TRUE;
    public static final int NOT_FACE = BinaryClassifier.FALSE;

    protected Vector<FaceSample> trainingData;

    public static void main(String[] args) throws Exception {

        System.out.println("Face Detection Problem. Version 1.0");

        if (args.length == 0) {
            System.out.println("Usage:");
            System.out.println("FaceDetectorProblem -t 30 -c Solution2 -x 10");
            System.out.println("t: Number of AdaBoost iterations.");
            System.out.println("c: Classname of outputted program.");
            System.out.println("x: Number of false cases to create per image.");
            System.exit(0);
        }

        int boostingSteps = 30;
        String className = "Solution2";
        int falseRatio = 10;

        for (int i = 0; i < args.length; i++) {
            String arg = args[i];
            if (arg.equals("-t")) {
                boostingSteps = Integer.parseInt(args[i + 1]);
                i++;
                continue;
            }
            if (arg.equals("-c")) {
                className = args[i + 1];
                i++;
                continue;
            }
            if (arg.equals("-x")) {
                falseRatio = Integer.parseInt(args[i + 1]);
                i++;
                continue;
            }
            System.out.println("Invalid argument: " + arg);
            System.exit(1);
        }

        System.out.println("Classname: " + className);
        System.out.println("Boosting steps: " + boostingSteps);
        System.out.println("False ratio: " + falseRatio);

        BoostingProblem p = new BoostingProblem(new FaceDetectorProblem1(falseRatio));

        p.boost(boostingSteps);

        AdaBoostSample[] samples = p.getSamples();

        p.save(className, "/home/ooechs/sxGA/src/problems/face/solutions/");

        p.toJava(className);

        p.test(samples);

    }

    public FaceDetectorProblem1(int falseRatio) {

        trainingData = new Vector<FaceSample>();

        final File trueDirectory2 = new File("/home/ooechs/ecj-training/faces/essex/mit/test/scaled32x40/");
        final File falseDirectory1 = new File("/home/ooechs/ecj-training/faces/essex/mit/false2/");
        final File falseDirectory2 = new File("/home/ooechs/ecj-training/faces/essex/false1/");

        addImageFiles(trueDirectory2, FACE, 500, falseRatio);
        addImageFiles(falseDirectory1, NOT_FACE, 500, falseRatio);
        addImageFiles(falseDirectory2, NOT_FACE, 500, falseRatio);

    }

    public AdaBoostSample[] getSamples() {
        AdaBoostSample[] samples = new AdaBoostSample[trainingData.size()];
        trainingData.toArray(samples);
        return samples;
    }

    public String getName() {
        return "Face Detector Problem";
    }

    public GAIndividual createIndividual() {
        return new FaceDetectorIndividual(0, 0, getFunctionID());
    }

    private BinaryClassifier calculateBinaryThreshold(GAIndividual ind) {

        StatisticsSolver faces = new StatisticsSolver();
        StatisticsSolver nonfaces = new StatisticsSolver();

        for (int i = 0; i < trainingData.size(); i++) {

            FaceSample data = trainingData.elementAt(i);

            // puts the window in the right place
            data.initHaar();

            // execute the individual
            double value = ind.execute(data.image);

            if (data.classID == FACE) {
                faces.addData(value);
            } else {
                nonfaces.addData(value);
            }

        }

        double faceAverage = faces.getMean();
        double nonFaceAverage = nonfaces.getMean();

        double threshold = (Math.abs(faceAverage - nonFaceAverage)  / 2) + Math.min(faceAverage, nonFaceAverage);
        if (faceAverage > nonFaceAverage) {
            return new BinaryClassifier(threshold, BinaryClassifier.OVER);
        } else {
            return new BinaryClassifier(threshold, BinaryClassifier.UNDER);
        }

    }

    public void evaluate(GAIndividual ind) {

        double TP = 0;
        double FP = 0;
        double TN = 0;
        double FN = 0;
        double totalTP = 0;
        int hits = 0;
        double error = 0;

        int returnedTrue = 0;
        int returnedFalse = 0;

        ind.setClassifier(calculateBinaryThreshold(ind));

        for (int i = 0; i < trainingData.size(); i++) {

            FaceSample faceSample = trainingData.elementAt(i);

            faceSample.initHaar();

            int classID = ind.classify(faceSample.image);

            if (faceSample.classID == FACE) {
                totalTP++;
                if (classID == faceSample.classID) {
                    // if it is a face and classifier was correct
                    TP += weights == null? 1 : weights[i];
                    hits++;
                    returnedTrue++;
                } else {
                    // if it is a face and classifier is incorrect
                    FN++;
                    returnedFalse++;
                }
            } else {
                if (classID == faceSample.classID) {
                    // if it isn't a face and classifier is correct
                    TN++;
                    hits++;
                    returnedFalse++;
                } else {
                    // if it isn't a face and classifier is incorrect.
                    FP += weights == null? 1 : weights[i];
                    returnedTrue++;
                }
            }

            if (classID != faceSample.classID) {
                error += weights == null? 1 : weights[i];
            }

        }

        if (returnedFalse == 0 || returnedTrue == 0) {
            // it has to get at least one of each type correct.
            ind.setWorstFitness();
        } else {

            //double alpha = 1;
            //double beta = 1;
            //double fitness = 1 / ((TP * alpha) / (totalTP + (FP * beta)));

            //ind.setKozaFitness(fitness);
            ind.setKozaFitness(error);
            ind.setHits(hits);

            //System.out.println("Hits: " + TP + ", " + TN);

        }

    }

    protected int addImageFiles(File directory, int classID, int max, int falseRatio) {
        if (!directory.exists()) throw new RuntimeException("Directory does not exist: " + directory.getAbsolutePath());
        int counter = 0;
        try {
            File[] trueFiles = directory.listFiles();

            for (int i = 0; i < trueFiles.length; i++) {
                File f = trueFiles[i];
                if (ImageFilenameFilter.isImage(f)) {

                    HaarRegions image = new HaarRegions(new PixelLoader(f));

                    for (int c = 0; c < falseRatio; c++) {
                        trainingData.add(new FaceSample(image, classID));
                        counter++;
                        // if (counter > max) break;
                        if (classID == FACE) break;                        
                    }

                }
            }
            System.out.println("\nLoaded: " + counter + " images.");

        } catch (Exception e) {
            e.printStackTrace();
        }
        return counter;
    }


    /**
     * There are four rectangle functions that we'd like to investigate.
     * @return
     */
    public int countFunctions() {
        return 4;
    }
}
